from flask import Flask, request, make_response
from kafka import KafkaConsumer
from redis import Redis
from time import sleep, time
import requests

from .helper import Helper


class Translator:

  def __init__(self, translation_function: callable) -> None:
      self._translation_function = translation_function

  def translate (self, line: str) -> list:
    return self._translation_function(line)

  def serve (self, host='localhost', port=5000, debug=False):
    app = Flask(__name__)

    @app.route('/metrics')
    def metrics ():
      data = request.get_json(True, True)

      if not data or 'metrics' not in data.keys():
        response = make_response('', 400)
      
      else:
        response = Helper.concatenate_metrics(list(map(self._translation_function, [ line for line in data['metrics'].split('\n') if line ])))
        response = make_response(response, 200)

      response.mimetype="text/plain"
      return response

    app.run(host=host, port=port, debug=debug)

  def redis_consume (self, metrics: list, redis_password: str):
    instance = metrics[0].properties['instance']

    conn = Redis(password=redis_password)
    conn.rpush(instance, metrics)
    conn.close()

  def prod (self, redis_password='root', kafka_host='localhost', kafka_port=9093, kafka_topic='general'):
    consumer = KafkaConsumer(kafka_topic, bootstrap_servers=[ '%s:%i' % (kafka_host, kafka_port) ], value_deserializer=lambda m: list(map(self._translation_function, m.decode('ascii').split('\n')) if m else []))
    
    for msg in consumer: 
      if msg.value: 
        self.redis_consume(msg.value, redis_password)

  def pull (self, endpoint: str, interval: int = 15, redis_password: str = 'root'):
    while True:
      ts = time()

      data = requests.get(endpoint).text
      metrics = self._translation_function(data)

      self.redis_consume(metrics, redis_password)

      time_left = interval - (time()-ts)
      if time_left > 0: sleep(time_left)



from sys import argv, stderr
from os import environ

from .helper import Helper
from .translator import Translator


class Runner: 
  CMD_TEXT = 'cmd'
  SERVER_TEXT = 'server'
  PROD_TEXT = 'prod'
  PULL_TEXT = 'pull'

  def __init__(self, translation_function: callable) -> None:
      self._translator = Translator(translation_function)

  def run (self):
    RUN_MODES = [ Runner.CMD_TEXT, Runner.SERVER_TEXT, Runner.PROD_TEXT, Runner.PULL_TEXT ]

    if len(argv) < 2 or argv[1] not in RUN_MODES:
      print("Run mode not recognised!\nUsage: python " + __file__ + " <run_mode> [OPTIONS]\nAvailable modes: " + ", ".join(RUN_MODES), file=stderr)

    elif argv[1] == Runner.CMD_TEXT:
      print(Helper.concatenate_metrics([self._translator.translate(a) for a in argv[2].split('\\n')]))

    elif argv[1] == Runner.SERVER_TEXT:
      args = {}

      flask_host = environ.get('FLASK_HOST')
      flask_port = environ.get('FLASK_PORT')
      flask_debug = environ.get('FLASK_DEBUG')

      if flask_host is not None: args['host'] = flask_host
      if flask_port is not None: args['port'] = flask_port
      if flask_debug is not None: args['debug'] = True

      self._translator.serve(**args)

    elif argv[1] == Runner.PROD_TEXT:
      args = {}

      redis_password = environ.get('REDIS_PASSWORD')
      kafka_host = environ.get('KAFKA_HOST')
      kafka_port = environ.get('KAFKA_PORT')
      kafka_topic = environ.get('KAFKA_TOPIC')

      if redis_password is not None: args['redis_password'] = redis_password
      if kafka_host is not None: args['kafka_host'] = kafka_host
      if kafka_port is not None: args['kafka_port'] = kafka_port
      if kafka_topic is not None: args['kafka_topic'] = kafka_topic

      self._translator.prod(**args)

    elif argv[1] == Runner.PULL_TEXT:
      args = {}

      pull_endpoint = environ.get('PULL_ENDPOINT')
      pull_interval = environ.get('PULL_INTERVAL')
      redis_password = environ.get('REDIS_PASSWORD')

      if pull_endpoint is not None: args['endpoint'] = pull_endpoint
      if pull_interval is not None: args['interval'] = int(pull_interval)
      if redis_password is not None: args['redis_password'] = redis_password

      self._translator.pull(**args)


class Helper:

  @staticmethod
  def concatenate_metrics (metrics: list) -> str:
    return '\n'.join([ Helper.concatenate_metrics(metric) if type(metric) == list else str(metric) for metric in metrics ])

  @staticmethod
  def concatenate_metrics_arrays (metrics: list) -> list:
    if metrics == []: return metrics
    if isinstance(metrics[0], list): return Helper.concatenate_metrics_arrays(metrics[0]) + Helper.concatenate_metrics_arrays(metrics[1:])
    return metrics[:1] + Helper.concatenate_metrics_arrays(metrics[1:])



from .metric import Metric


class MetricsFactory:

  @classmethod
  def create_load_metric (cls, interval: int, value: str, ts: str, instance: str):
    return Metric('node_load' + str(interval), value, { }, ts, instance)

  @classmethod
  def create_cpu_time_metric (cls, cpu: str, mode: str, value: str, ts: str, instance: str):
    return Metric('node_cpu_seconds_total', value, { 'cpu': cpu, 'mode': mode }, ts, instance)

  @classmethod
  def create_filesystem_metric (cls, title: str, device: str, fstype: str, mountpoint: str, value: str, ts: str, instance: str):
    return Metric('node_filesystem_%s_bytes' % (title,), value, { 'device': device, 'fstype': fstype, 'mountpoint': mountpoint }, ts, instance)

  @classmethod
  def create_filesystem_available_metric (cls, device: str, fstype: str, mountpoint: str, value: str, ts: str, instance: str):
    return cls.create_filesystem_metric('avail', device, fstype, mountpoint, value, ts, instance)

  @classmethod
  def create_filesystem_size_metric (cls, device: str, fstype: str, mountpoint: str, value: str, ts: str, instance: str):
    return cls.create_filesystem_metric('size', device, fstype, mountpoint, value, ts, instance)

  @classmethod
  def create_memory_metric (cls, title: str, value: str, ts: str, instance: str):
    return Metric('node_memory_%s_bytes' % (title,), value, {  }, ts, instance)

  @classmethod
  def create_memory_total_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('MemTotal', value, ts, instance)

  @classmethod
  def create_memory_free_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('MemFree', value, ts, instance)

  @classmethod
  def create_memory_available_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('MemAvailable', value, ts, instance)

  @classmethod
  def create_memory_cached_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Cached', value, ts, instance)

  @classmethod
  def create_memory_buffered_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Buffers', value, ts, instance)

  @classmethod
  def create_memory_slab_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Slab', value, ts, instance)

  @classmethod
  def create_memory_page_tables_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('PageTables', value, ts, instance)

  @classmethod
  def create_memory_inactive_metric (cls, value: str, ts: str, instance: str, detail: str = None):
    return cls.create_memory_metric('Inactive%s' % ('' if detail is None else '_' + detail,), value, ts, instance)

  @classmethod
  def create_memory_active_metric (cls, value: str, ts: str, instance: str, detail: str = None):
    return cls.create_memory_metric('Active%s' % ('' if detail is None else '_' + detail,), value, ts, instance)

  @classmethod
  def create_memory_committed_as_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Committed_AS', value, ts, instance)

  @classmethod
  def create_memory_commit_limit_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('CommitLimit', value, ts, instance)

  @classmethod
  def create_memory_swap_metric (cls, title: str, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Swap%s' % (title,), value, ts, instance)

  @classmethod
  def create_memory_swap_total_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_swap_metric('Total', value, ts, instance)

  @classmethod
  def create_memory_swap_free_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_swap_metric('Free', value, ts, instance)

  @classmethod
  def create_memory_vmalloc_metric (cls, title: str, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Vmalloc%s' % (title,), value, ts, instance)

  @classmethod
  def create_memory_vmalloc_chunk_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_vmalloc_metric('Chunk', value, ts, instance)

  @classmethod
  def create_memory_vmalloc_total_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_vmalloc_metric('Total', value, ts, instance)

  @classmethod
  def create_memory_vmalloc_used_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_vmalloc_metric('Used', value, ts, instance)

  @classmethod
  def create_memory_writeback_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Writeback', value, ts, instance)

  @classmethod
  def create_memory_writeback_tmp_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('WritebackTmp', value, ts, instance)

  @classmethod
  def create_memory_dirty_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Dirty', value, ts, instance)

  @classmethod
  def create_memory_mapped_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Mapped', value, ts, instance)

  @classmethod
  def create_memory_reclaimable_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('SReclaimable', value, ts, instance)

  @classmethod
  def create_memory_huge_pages_metric (cls, value: str, ts: str, instance: str, detail: str = None):
    return Metric('node_memory_HugePages_%s' % ('' if detail is None else '_' + detail,), value, {}, ts, instance)

  @classmethod
  def create_memory_huge_pages_free_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_huge_pages_metric(value, ts, instance, 'Free')

  @classmethod
  def create_memory_huge_pages_rsvd_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_huge_pages_metric(value, ts, instance, 'Rsvd')

  @classmethod
  def create_memory_huge_pages_total_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_huge_pages_metric(value, ts, instance, 'Total')

  @classmethod
  def create_memory_direct_map_metric (cls, value: str, ts: str, instance: str, detail: str = None):
    return cls.create_memory_metric('DirectMap%s' % ('' if detail is None else '_' + detail,), value, ts, instance)

  @classmethod
  def create_memory_direct_map_1g_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_direct_map_metric(value, ts, instance, '1G')

  @classmethod
  def create_memory_direct_map_2m_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_direct_map_metric(value, ts, instance, '2M')

  @classmethod
  def create_memory_direct_map_4k_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_direct_map_metric(value, ts, instance, '4k')

  @classmethod
  def create_memory_huge_pages_free_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_huge_pages_metric(value, ts, instance, 'Free')

  @classmethod
  def create_memory_huge_pages_rsvd_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_huge_pages_metric(value, ts, instance, 'Rsvd')

  @classmethod
  def create_memory_huge_pages_total_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_huge_pages_metric(value, ts, instance, 'Total')

  @classmethod
  def create_memory_huge_pages_size_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Hugepagesize', value, ts, instance)

  @classmethod
  def create_memory_unevictable_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Unevictable', value, ts, instance)

  @classmethod
  def create_memory_mlocked_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('Mlocked', value, ts, instance)

  @classmethod
  def create_memory_nfs_metric (cls, value: str, ts: str, instance: str):
    return cls.create_memory_metric('NFS_Unstable', value, ts, instance)

  @classmethod
  def create_up_time_metric (cls, value: str, ts: str, instance: str):
    return Metric('node_time_seconds', value, {  }, ts, instance)

  @classmethod
  def create_boot_time_metric (cls, value: str, ts: str, instance: str):
    return Metric('node_boot_time_seconds', value, {  }, ts, instance)

  @classmethod
  def create_network_bytes_metric (cls, title: str, value: str, ts: str, instance: str):
    return Metric('node_network_%s_bytes_total' % (title,), value, {  }, ts, instance)

  @classmethod
  def create_network_receive_bytes_metric (cls, value: str, ts: str, instance: str):
    return cls.create_network_bytes_metric('receive', value, ts, instance)

  @classmethod
  def create_network_transmit_bytes_metric (cls, value: str, ts: str, instance: str):
    return cls.create_network_bytes_metric('transmit', value, ts, instance)

  @classmethod
  def create_disk_metric (cls, title: str, device: str, value: str, ts: str, instance: str):
    return Metric('node_disk_%s_total' % (title,), value, { 'device': device }, ts, instance)

  @classmethod
  def create_disk_completed_metric (cls, title: str, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('%s_completed' % (title,), device, value, ts, instance)

  @classmethod
  def create_disk_reads_completed_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_completed_metric('reads', device, value, ts, instance)

  @classmethod
  def create_disk_writes_completed_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_completed_metric('writes', device, value, ts, instance)

  @classmethod
  def create_disk_bytes_metric (cls, title: str, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('%s_bytes' % (title,), device, value, ts, instance)

  @classmethod
  def create_disk_read_bytes_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_bytes_metric('read', device, value, ts, instance)

  @classmethod
  def create_disk_written_bytes_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_bytes_metric('written', device, value, ts, instance)

  @classmethod
  def create_disk_io_time_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('io_time_seconds', device, value, ts, instance)

  @classmethod
  def create_disk_io_time_weighted_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('io_time_weighted_seconds', device, value, ts, instance)

  @classmethod
  def create_disk_io_read_time_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('read_time_seconds', device, value, ts, instance)

  @classmethod
  def create_disk_io_reads_merged_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('reads_merged', device, value, ts, instance)

  @classmethod
  def create_disk_io_writes_merged_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('writes_merged', device, value, ts, instance)

  @classmethod
  def create_disk_io_write_time_metric (cls, device: str, value: str, ts: str, instance: str):
    return cls.create_disk_metric('write_time_seconds', device, value, ts, instance)

  @classmethod
  def create_disk_io_now_metric (cls, device: str, value: str, ts: str, instance: str):
    return Metric('node_disk_io_now', value, { 'device': device }, ts, instance)

  @classmethod
  def create_vmstat_metric (cls, title: str, value: str, ts: str, instance: str):
    return Metric('node_vmstat_%s' % (title,), value, { }, ts, instance)



from time import time

from .consts import TIMESTAMP_LENGTH


class Metric:

  def __init__ (self, title: str, value: str, properties: dict = None, timestamp: float = None, instance: str = ''):
    self._title = title
    self._value = value
    self._properties = properties if properties is not None else {}
    self._timestamp = time() if timestamp is None else timestamp
    
    if instance is not None: self._properties['instance'] = instance

    self._str_cache = None

  @property
  def title (self): 
    return self._title

  @property
  def value (self): 
    return self._value

  @property
  def properties (self): 
    return self._properties

  @property
  def timestamp (self): 
    return self._timestamp

  def __str__(self) -> str:
    if self._str_cache is None:
      value = self.value.replace('i', '')

      properties = ','.join(['%s="%s"' % (prop, value) for prop, value in self.properties.items()])

      timestamp = str(self.timestamp)[:TIMESTAMP_LENGTH]
      timestamp += ''.join(['0' for _ in range(TIMESTAMP_LENGTH-len(timestamp))])

      self._str_cache = "%s{%s} %s %s" % (self._title, properties, value, timestamp)

    return self._str_cache
